import numpy as np
import time

class DummyHeater():
    def __init__(self, address):
        self.address = address
        self.voltage = 0

    def set_voltage(self, voltage):
        self.voltage = voltage
        print(f'Dummy heater at address {self.address} voltage now {voltage}')

    def get_voltage(self):
        return self.voltage

    def get_current(self):
        return round(time.time() % 10, 2)
    

class DummyThermometer():
    def __init__(self, address):
        self.address = address

    def read_temperature(self):
        return round(time.time() % 2, 2) + np.random.randn()*0.1
    
    
class DummyRelay():
    def __init__(self, address):
        self.address = address
        self.state = False

    def set_state(self, state):
        self.state = state
        print(f'Dummy relay at address {self.address} state now {state}')
    
    def get_state(self):
        return self.state


# %%
from statemachine import State, StateMachine
import time
import tomllib
from simple_pid import PID

class Single_shot_1k(StateMachine):
    def __init__(self, settings_toml, hal_client, monitor_client):
        # Add ZMQ connection - HAL Client
        self.hal_client = hal_client
        # Add html scrapper - Monitor Client
        self.monitor_client = monitor_client
        # Import settings from toml file
        with open(f'{settings_toml}', "rb") as f:
            self.sm_settings = tomllib.load(f)
        
        # Keep track of time when we enter states. It is set on_entry into new state
        self.enter_state_time = 0
        #Check if just entered state
        self.just_entered = True
        # Controls if the state machine is running (Must call start_state_machine() )
        self.cycle_running = False
        # Time.sleep before state is ran again
        self.state_machine_cycle_time = self.sm_settings['misc']['state_machine_cycle_time'] #seconds
        # Holds all fridge data from metric server. refreshed every state_machine_cycle_time
        self.fridge_state = {}
        self.fridge_timer = ElapsedTime()
        self.dt = 0
        # Clock recycle bool to sotp pre_cool from going bac into head_cold
        # I dont think This ever needs to be false other than first start up...
        self.clock_recycle = False
        # PID controller setup
        self.pid = {}
        # Get max values of HAL TOML hardware by grabbing metric server data
        self.update_fridge_state()
        self.setup_pids()
        self.currently_in_use_pids = {}
        self.time_elapse_dict = {}
        super().__init__()
        # Allows states to move on without transition (required dont touch)
        self.allow_event_without_transition = 1

    def setup_pids(self):
        for key,values in self.sm_settings['pid_configuration'].items():
            max_value = self.fridge_state['heater_max_values'][key]
            self.pid[key] = PID(Kp = values['kp'], 
                                        Kd = values['kd'],
                                        Ki = values['ki'],
                                        sample_time=None,
                                        output_limits = (0, max_value))
        
    #-----------------------------------------------------------------------------------------------------#
    #                                   functions used for grabbing data                                  #
    #-----------------------------------------------------------------------------------------------------#

    # This grabs values of nested sensor_keys from a specific dictionary key
    def get_values_from_keys(self,key, sensor_keys, data_json):
        results = {}
        for sensor in sensor_keys:
            if sensor in data_json[f'{key}']:
                results[sensor] = data_json[f'{key}'][sensor]
        return results
    
    # This function will call the query metric server and return only the temperature of the sensors
    def get_temperature(self, sensors):
        data = self.monitor_client.get_metrics()
        return self.get_values_from_keys(sensor_keys = sensors, data_json = data, key = 'temperatures')
    
    # This function will call the query metric server and return only the heater values
    def get_heater_value(self, heater):
        data = self.monitor_client.get_metrics()
        return self.get_values_from_keys(sensor_keys= heater, data_json = data, key = 'heaters')

    #-----------------------------------------------------------------------------------------------------#
    #                      Action functions used to communicate with HAL.                                 #
    #-----------------------------------------------------------------------------------------------------#
    
    def set_heater_value(self, heater, value):
        response = self.hal_client.set_heater_value(f'{heater}', f'{value}')
        return response

    #-----------------------------------------------------------------------------------------------------#
    # Non-Async functions for starting the state machine & main cycle
    #------------------------------------------------------------------------------------------------------#
    def turn_on_state_machine(self):
        # Starts state machine cycle
        self.cycle_running = True
        self.run_cycle()
    def turn_off_state_machine(self):    
        # Stops state machine cycle
        self.cycle_running = False
    def run_cycle(self):
            while self.cycle_running:
                # Update the temperatures inside the state machine
                self.update_fridge_state()
                # Run the state machine to check if we should move to the next state or not
                # If we move to a new state, update the list of currently-in-use PIDs
                self.cycle()
                # Scan through the list of currently-in-use PIDs and update their setpoints
                self.update_pid_values()
                time.sleep(self.state_machine_cycle_time)
    #------------------------------------------------------------------------------------------------------#
    #                                       Misc functions for the state machine
    #------------------------------------------------------------------------------------------------------#
    def update_fridge_state(self):
        fridge_data = self.monitor_client.get_metrics()
        self.fridge_state = fridge_data
        self.dt = self.fridge_timer()[1]
        self.fridge_timer.reset()

    def update_pid_values(self):
    # Update the setpoints of all PIDs in the list of currently-in-use PIDs
        #print('Updating PID values and sending to HAL')
        for pid_key in self.currently_in_use_pids:
            # Get current temperature of key
            current_value = self.fridge_state['temperatures'][pid_key]
            # Define state machine setpoint of key Ex 4PUMP @ 45K
            setpoint = self.currently_in_use_pids[pid_key]
            # Update PID setpoint
            self.pid[pid_key].setpoint = setpoint
            # Get new value from PID
            new_value = self.pid[pid_key](input_ = current_value, dt = self.dt)
            # Send new value to HAL
            #print('Current Temp =',current_value ,'Setting', pid_key, 'to', new_value, 'with dt =', self.dt)
            self.set_heater_value(heater = pid_key, value = new_value)

    # Checks temperature of fridge and returns True if all temperatures are below the threshold
    def check_temperature_criteria(self, state, excluded_variable = None):
        threshold_temps = self.sm_settings[f'{state}']['temperature_thresholds']
        current_temps = self.fridge_state['temperatures']
        # Filters current_temps to only include keys that are in threshold_temps
        filtered_current_data = {key: current_temps[key] for key in threshold_temps if key in current_temps}
        # Exclude a specific variable from the check
        if all(filtered_current_data[thermometer] < threshold_temps[thermometer] 
               for thermometer in threshold_temps if thermometer != excluded_variable):
            return True
        else:
            return False
    def check_temperature_criteria_sensor(self, sensor):
        # Get the temperature of the sensor
        current_temp = self.fridge_state['temperatures'][sensor]
        # Get the threshold for the sensor
        state = self.current_state.id
        threshold_temp = self.sm_settings[f'{state}']['temperature_thresholds'][sensor]
        # Check if the current temperature is below the threshold
        if current_temp < threshold_temp:
            return True
        else:
            return False
    # This will find the heaters and switches that contain 'pid_set_value' in their subkey. 
    # It is only called on entry to make self.currently_in_use_pids
    def find_sub_subkey(self, state, sub_subkey):
        """
        Find all subkeys and their values where the given sub-subkey exists in a nested dictionary.

        Parameters:
        - data (dict): The main dictionary with nested structures.
        - sub_subkey (str): The sub-subkey to search for.

        Returns:
        - dict: A dictionary of subkeys and their values where the sub-subkey exists.
        """
        data = self.sm_settings[f'{state}']
        result = {}

        for key, subkeys in data.items():
            # Ensure subkeys is a dictionary
            if isinstance(subkeys, dict):
                for subkey, subvalue in subkeys.items():
                    # Check if the subvalue is a dictionary and contains the sub-subkey
                    if isinstance(subvalue, dict) and sub_subkey in subvalue:
                        result[subkey] = subvalue[sub_subkey]

        return result
    #Makes timer objects
    def make_timers(self, dict_of_timers):
        for key, value in dict_of_timers.items():
            self.time_elapse_dict[key] = ElapsedTime(value)
            
    #Function for cycling through timers and checking if true
    def timer_check(self):
        for key in self.time_elapse_dict:
            #Check if time passed is true
            #EDIT ME I remove [0]
            if self.time_elapse_dict[key]()[0]:
                return True
        return False

    # Function used for keeping track of time in the state
    # EDIT ME Add console / logging stuff
    def check_just_entered(self):
        if self.just_entered:
            self.just_entered = False
            self.enterStateTime = time.time()
            #Console / logging stuff
            return True
        else:
            return False
    def turn_off_not_in_use_pids(self):
        for key in self.pid:
            if key not in self.currently_in_use_pids.keys():
                print('Turning off', key)
                self.set_heater_value(key, 0)

    def run_per_state(self):
        # Resets time_elapse_dict
        self.time_elapse_dict = {}
        # Resets currently in use PIDs
        self.currently_in_use_pids = {}
        # Rebuilds currently in use PIDs from the state machine settings
        self.currently_in_use_pids = self.find_sub_subkey(self.current_state.id, sub_subkey = 'pid_set_value')
        # Sets up elapse time objects
        self.dict_to_turn_into_timers = self.find_sub_subkey(self.current_state.id, sub_subkey = 'elapse_time_max')
        self.make_timers(self.dict_to_turn_into_timers)


        # EDIT ME 
        self.check_just_entered
        self.turn_off_not_in_use_pids()

    def log_time(self):
        date = time.localtime()
        log_time = f'{date.tm_year}-{date.tm_mon}-{date.tm_mday} {date.tm_hour}:{date.tm_min}:{date.tm_sec}'
        return log_time
    #-----------------------------------------------------------------------------------------------------#
    #                                    Defining states of state machine
    #-----------------------------------------------------------------------------------------------------#
    # Wait for the cryocooler to cool down 4K to below 4.1K
    pre_cool = State(name = 'Precooling 4K to below 4.1K' , initial = True, value = 1)
    #-----------------------------------------------------------------------------------------------------#
    # Prepare cycle by turning off pumps and switches
    # Wait for the compressor to cool the heat switch to below < 10K before starting state machine
    start_recycle = State(name = 'Turning off switches & pumps and waiting for 4heat switch too cool below 10K', value = 2)
    #-----------------------------------------------------------------------------------------------------#
    # I believe this state was added to not thermally shock the system when heating to 45K. 
    get_pump_warm = State(name = 'Warming 4pump to 25K to not thermal shock system', value = 3)
    #-----------------------------------------------------------------------------------------------------#
    # Turn on 4 pumps for a period of time "30*60*1000" or if 4head < 3K #EDIT ME -> change time 
    get_pump_hot = State(name = 'warming 4pump to 45K', value = 4)
    #-----------------------------------------------------------------------------------------------------#
    # Turn off 4pump 
    turn_off_pump = State(name = 'Turning off 4pump', value = 5)
    #-----------------------------------------------------------------------------------------------------#
    # Turn on heat switch 4 to cool down 4 head, 
    turn_on_heat_switch_4 = State(name = 'Turning on 4switch and cooling down', value = 6)
    #-----------------------------------------------------------------------------------------------------#
    # Confirm state of 4head being <1K
    head_cold = State(name = '4head is below 1K', value =7)
    #-----------------------------------------------------------------------------------------------------#
    # To speed up system warm up, turning on the heat switches to thermally connect the whole system & HPHs
    # Also turn off still heater? maybe keep it on to warm up still faster... It jsut wont cycle heilum bc not super fluid
    warm_up_system = State(name = 'Warming up system', value = 8, final = True)
    #-----------------------------------------------------------------------------------------------------#
    
    #-----------------------------------------------------------------------------------------------------#
    #                               Defining transitions between s•••••••••••tates
    #-----------------------------------------------------------------------------------------------------#
    # Normal 1K cycle between states
    cycle = (
        pre_cool.to(head_cold, cond = 'already_cold_cond')                |
        pre_cool.to(start_recycle, cond = 'pre_cool_cond')                | 
        start_recycle.to(get_pump_warm, cond = 'start_recycle_cond')      |
        get_pump_warm.to(get_pump_hot, cond = 'pump_warm_cond')           |
        get_pump_hot.to(turn_off_pump, cond = 'pump_hot_cond' )           |
        turn_off_pump.to(turn_on_heat_switch_4, cond = 'pump_off_cond')   |
        turn_on_heat_switch_4.to(head_cold, cond = 'heat_switch_on_cond') |
        turn_on_heat_switch_4.to(pre_cool, cond = 'head_not_cooling_cond')|
        head_cold.to(pre_cool, cond = 'head_cold_recycle_cond')           |
        head_cold.to(warm_up_system, cond = 'stop_func')                             
    )    
    #-----------------------------------------------------------------------------------------------------#
    #        Condition Functions - These determine if the state can transition to the next State          #
    #-----------------------------------------------------------------------------------------------------#
    def pre_cool_cond(self):
        # Goals
        # Check if 4K stage is < 4.1K
        #print('checking mainplate', flush = True)
        return self.check_temperature_criteria(self.current_state.id, excluded_variable= '4HEAD')    
    def already_cold_cond(self):
        # Check if 1K head is already cold
        #print('checking 4head', flush = True)
        if self.clock_recycle == False:
            return self.check_temperature_criteria_sensor(sensor = '4HEAD')
    def start_recycle_cond(self):
        # Goals
        # Make sure pump and switch is off before starting cycle
        # Check if heat switch is less than 10K 
        return self.check_temperature_criteria(self.current_state.id)       
    def pump_warm_cond(self):
        # PD warming slowly to 45K to thermall shock system (too much heat too fast causing overshoot)
        return self.timer_check()
    def pump_hot_cond(self):
        # PD controling at 45K. PD called on entery of state
        # Once 4head is below 3.8K, move to next state
        if self.check_temperature_criteria(self.current_state.id):
            return True
        else:
            return self.timer_check()
    def pump_off_cond(self):
        return self.timer_check()
    def heat_switch_on_cond(self):
        # Move to head_cold state if below 875mK
        return self.check_temperature_criteria(self.current_state.id)
    def head_not_cooling_cond(self):
        # If 4head doesnt get below 0.875K in 1.5 hours, Try recycling
        return self.timer_check()
    def head_cold_recycle_cond(self):
        # Initialize a timer when the cold_stage_temp exceeds the threshold
        if not hasattr(self, 'head_cold_timer'):
            self.head_cold_timer = ElapsedTime(threshold_seconds=10)

        # Get current time and temperature
        current_hour = time.localtime().tm_hour
        current_min = time.localtime().tm_min
        target_hour = self.sm_settings['misc']['start_recycle_clock_hour']
        target_min = self.sm_settings['misc']['start_recycle_clock_min']
        # Get the current temperature of the 4HEAD
        cold_stage_temp = self.fridge_state['temperatures']['4HEAD']
        threshold = self.sm_settings['misc']['start_recycle_temp']

        # Check if 4HEAD temperature exceeds the threshold
        if cold_stage_temp > threshold:
            print(f'4Head is starting to warm up > {threshold}K, monitoring timer...')
            if self.head_cold_timer()[0]:  # Check if timer exceeds 10 seconds
                print(f'4Head has been above {threshold}K for 10 seconds, will begin recycle', f'{self.log_time()}')
                return True
        else:
            # Reset the timer if the temperature drops below the threshold
            self.head_cold_timer.reset()

        # Check if it's the target time to start recycling
        if current_hour == target_hour and current_min == target_min:
            print(f'It is {target_hour}:{target_min}, will begin recycle')
            self.clock_recycle = True
            return True

        return False
    def stop_func(self, bool_test = False):
        #-----------------------------------------------------------------------------------------------------#
        # Exit condition used for all states to end the state machine
        #-----------------------------------------------------------------------------------------------------#
        # EDIT ME
        #-----------------------------------------------------------------------------------------------------#
        return bool_test
    #-----------------------------------------------------------------------------------------------------#
    #        On-entry functions - These are the functions/tasks that are executed on entry into a state.  #
    #-----------------------------------------------------------------------------------------------------#    
    def on_enter_pre_cool(self):

        # Turn off heaters and switches during precooling
        print('on_enter_pre_cool', f'{self.log_time()}')
        self.run_per_state()
    def on_enter_start_recycle(self):
        print('on_enter_start_recycle', f'{self.log_time()}')
        self.run_per_state()
    def on_enter_get_pump_warm(self):
        self.run_per_state()
        print('on_enter_pump_warm', f'{self.log_time()}')
    def on_enter_get_pump_hot(self):
        print('on_enter_get_pump_hot', f'{self.log_time()}')
        self.run_per_state()

        #self.set_all_key_values(self.current_state.id, 'relays')
    def on_enter_turn_on_heat_switch_4(self):
        self.run_per_state()
        print('Turning on 4switch to begin <1K cooldown', f'{self.log_time()}', flush=True)
    def on_enter_head_cold(self):
        self.run_per_state()
        #self.set_all_key_values(self.current_state.id, 'relays')
        print('Head is below 875mK', f'{self.log_time()}', flush=True)
class ElapsedTime:
    def __init__(self, threshold_seconds=None):
        """
        Initialize the ElapsedTime object.
    

        Parameters:
        - threshold_seconds (float, optional): The time threshold in seconds.
          If None, threshold checks will not be performed.
        """
        self.threshold_seconds = threshold_seconds
        self.start_time = time.time()

    def reset(self):
        """
        Reset the timer.
        """
        self.start_time = time.time()

    def __call__(self):
        """
        Check if the elapsed time exceeds the optional threshold and return elapsed time.

        Returns:
        - (bool or None, float): A tuple containing:
            - bool: True if elapsed time exceeds the threshold, False otherwise. 
                    If no threshold is provided, this is None.
            - float: Elapsed time in seconds since the timer was started.
        """
        elapsed_time = time.time() - self.start_time

        # Check if threshold_seconds is provided
        if self.threshold_seconds is None:
            return None, elapsed_time
        
        return elapsed_time >= self.threshold_seconds, elapsed_time

    def time_remaining(self):
        """
        Get the remaining time until the threshold is reached.

        Returns:
        - float or None: Time remaining in seconds. 
                         If the threshold is exceeded, returns 0.
                         If no threshold is set, returns None.
        """
        if self.threshold_seconds is None:
            return None

        elapsed = time.time() - self.start_time
        return max(0, self.threshold_seconds - elapsed)





import serial

class KD3005P(object):
    """Python class for KORAD KD3005P single channel programmable power supply.
    Rated for 30V @ 5A. """
    #Site: https://www.koradtechnology.com/product/84.html
    #Program Manual: https://www.sra-shops.com/pub/media/docs/srasolder/instructions/kd3005p-user-manual.pdf
    
    
    def __init__(self, serial_address):
        self.serial = serial.Serial(serial_address, timeout = 1, baudrate = 9600)

    def read(self):
        return self.serial.readline().strip().decode('utf-8')

    def write(self, string):
        write_string = string + '\n'
        self.serial.write(write_string.encode('utf-8'))

    def query(self, string):
        self.write(string)
        return self.read()

    def reset(self):
        self.write('*RST')

    def identify(self):
        return self.write('*IDN?')

    def close(self):
        return self.serial.close()
    

    def set_current(self, amps, channel = 1):
        #Sets output current. Input must be an int or a float
        msg = f'ISET{channel}:{amps}'
        self.write(msg)
            
    def set_voltage(self, volts, channel = 1):
        #Sets output voltage. Input must be an int or a float
        self.write(f'VSET{channel}:{volts}')
            
    def read_current_setting(self, channel = 1):
        #Returns the output current setting
        current = self.query(f'ISET{channel}?')
        return float(current)
    
    def read_voltage_setting(self, channel = 1):
        #Returns the output current setting
        voltage = self.query(f'VSET{channel}?')
        return float(voltage)
    
    def read_current(self, channel = 1):
        #Returns actual output current
        current = self.query(f'IOUT{channel}?')
        return float(current)

    def read_voltage(self, channel = 1):
        #Returns actual output voltage
        voltage = self.query(f'VOUT{channel}?')
        return float(voltage)


# p = KD3005P('/dev/ttyACM0')
# print(p.read_current())


import serial
import numpy as np

class SIM921:
    """Python class for SRS SIM921 AC resistance bridge inside a SIM900
    mainframe, written by Adam McCaughan"""
    def __init__(self, serial_address, sim900port):
        self.serial = serial.Serial(serial_address, timeout = 1, baudrate = 115200)
        self.sim900port = sim900port
        self.serial.send_break()

    def read(self):
        return self.serial.readline().strip().decode('utf-8')

    def write(self, string):
        write_string = string + '\r\n'
        self.serial.write(write_string.encode('utf-8'))

    def query(self, string):
        self.write(string)
        return self.read()

    def reset(self):
        self.write_simport('*RST')

    def identify(self):
        return self.query_simport('*IDN?')

    def close(self):
        return self.serial.close()

    def write_simport(self, message):
        self.write('CONN %d,"xyz"' % self.sim900port)
        self.write(message)
        self.write('xyz')

    def read_simport(self):
        self.write('CONN %d,"xyz"' % self.sim900port)
        data = self.read()
        self.write('xyz')
        return data

    def query_simport(self, message):
        self.write('CONN %d,"xyz"' % self.sim900port)
        reply = self.query(message)
        self.write('xyz')
        return reply

    def read_resistance(self):
        R = float(self.query_simport('RVAL?'))
        return R

    def read_temperature(self):
        T = float(self.query_simport('TVAL?'))
        return T

    def set_range(self, max_resistance = 200):
        range_code = np.ceil(np.log10(max_resistance/(20e-3)))
        range_code = int(range_code)
        self.write_simport('RANG %s' % range_code)

    def set_excitation(self, voltage = 10e-6):
        valid_voltages = [3e-6, 10e-6, 30e-6, 100e-6, 300e-6, 1e-3, 3e-3, 10e-3, 30e-3]
        if voltage not in valid_voltages:
            raise ValueError('Excitation voltage must be in [3e-6, 10e-6, 30e-6, 100e-6, 300e-6, 1e-3, 3e-3, 10e-3, 30e-3]')
        excitation_code = np.where(voltage == np.array(valid_voltages))[0][0]
        self.write_simport('EXCI %s' % excitation_code)

    def set_time_constant(self, time_constant = 1):
        valid_tc = [0.3, 1, 3, 10, 30, 100, 300]
        if time_constant not in valid_tc:
            raise ValueError('Time constant must be in [0.3, 1, 3, 10, 30, 100, 300]')
        time_constant_code = np.where(time_constant == np.array(valid_tc))[0][0]
        self.write_simport('TCON %s' % time_constant_code)

# s = SIM921('/dev/ttyUSB0', 2)
# print(s.identify())


import zmqhelper as zmqh
import json


class KnockKnockServer(zmqh.Server):
    def __init__(self, port, n_workers, server_name='Zelda'):
        self.server_name = server_name
        self.cmds = [('knock','Knock knock joke'), 
            ('name', 'Server name'), 
            ('all', 'All available commands')]
        super().__init__(port, n_workers)

    def handle(self,msg):
        inputs = json.loads(msg)
        cmd = inputs['cmd']
        query = inputs['query']
        cmd = cmd.lower()

        try:
            if cmd == "knock":
                msg_out = self.server_name+' is here!'

            elif cmd == 'name':
                msg_out = self.server_name
            elif cmd == 'all':
                msg_out = self.get_all_commands()
            else:
                msg_out = 'Invalid command, valid commands are: '
                msg_out += self.get_all_commands()

            # Catch errors and return them
        except Exception as e:
            msg_out = "Error: %r" % e
            raise e
        msg_out = msg_out.encode('utf-8')
        return msg_out

    def get_all_commands(self):
        msg_out = ''
        for el in self.cmds:
            msg_out+= el[0]+': '+el[1]+ ', '
        msg_out = msg_out[0:-2] # remove trailing comma
        return msg_out

if __name__ == '__main__':
    print('Starting Knock Knock Server')
    my_server = KnockKnockServer(port='5553', n_workers=1, server_name='Link')


from setuptools import setup

setup(name='zmqhelper',
      version='0.1',
      description='Convenient server and client classes for pyZMQ',
      url='',
      author='Krister Shalm, Joseph Cavanagh',
      author_email='lks@nist.gov',
      license='MIT',
      packages=['zmqhelper'],
      include_package_data=True,
      zip_safe=False)



# All classes in this file are thin wrappers for the main driver classes in the other files
# that allow the HAL to interact with the classes in a consistent way.

# Each thermometer class must implement the following methods:
    # setup(...)
    # get_temperature()
# And each heater class must implement the following methods:
    # setup(...)
    # set_heater_value(value)
    # get_heater_value()
# If you add a new driver, make sure to add it to the hal_classes dictionary at the bottom of this file.

from fridgeos.hal.drivers.korad_kd3005p import KD3005P
from fridgeos.hal.drivers.srs_sim921 import SIM921
from fridgeos.hal.drivers.srs_sim922 import SIM922
from fridgeos.hal.drivers.swarm import Lockin, Diode, PumpHeater, HeatSwitch, WarmupHeater

### HEATERS

class HAL_KD3005P():
    def setup(self, address):
        self.heater = KD3005P(address)
    
    def set_heater_value(self, value):
        self.heater.set_voltage(value)
    
    def get_heater_value(self):
        return self.heater.read_voltage()
    
class SWARM_HPH():
    def setup(self, address, mux_name = None):
        self.heater = PumpHeater(address, mux_name)

    def set_heater_value(self, value):
        self.heater.set_pump_current(value)

    def get_heater_value(self):
        return self.heater.get_pump_measurement()
    
class SWARM_LPH():
    def setup(self, address, mux_name):
        self.heater = HeatSwitch(address, mux_name)

    def set_heater_value(self, value):
        self.heater.set_heat_switch_voltage(value)

    def set_heater_enable(self, enable):
        self.heater.set_heat_switch_enable(enable)

    def get_heater_enable(self):
        return self.heater.get_heat_switch_enable()
    
    def get_heater_value(self):
        return self.heater.get_heat_switch_voltage()


### THERMOMETERS 

class HAL_SIM921():
    def setup(self, address, slot):
        self.thermometer = SIM921(address, sim900port=slot)
    
    def get_temperature(self):
        return self.thermometer.read_temperature()
    
class HAL_SIM922():
    def setup(self, address, slot, channel):
        self.thermometer = SIM922(address, sim900port=slot, channel=channel)
    
    def get_temperature(self):
        return self.thermometer.read_temperature()
    
class SWARM_LOCKIN():
    def setup(self, address, calibration_file = None, mux_name = None, mux = False):
        self.thermometer = Lockin(address, calibration_file, mux_name, mux)

    def get_temperature(self):
        return self.thermometer.read_temp()
    
class SWARM_DIODE():
    def setup(self, address, calibration_file = None, mux_name = None):
        self.thermometer = Diode(address, calibration_file, mux_name)

    def get_temperature(self):
        return self.thermometer.read_temp()


hal_classes = {
    'korad-kd3005p': HAL_KD3005P,
    'srs-sim921': HAL_SIM921,
    'srs-sim922': HAL_SIM922,
    'swarm_lockin': SWARM_LOCKIN,
    'swarm_diode': SWARM_DIODE,
    'swarm_hph': SWARM_HPH,
    'swarm_lph': SWARM_LPH
}


#%%
import numpy as np
from serial import Serial
import json

class Lockin():
    def __init__(self, address, calibration_file = None, name = None, mux = False):
        try:
            self.serial = Serial(address)
            self.serial.close()
        except Exception as e:
            print(e)
            pass
        if calibration_file != None:
            self.calibration_array = np.loadtxt(calibration_file, delimiter =',')
        else:
            self.calibration_array = None
        self.name = name
        self.mux = mux
    def read(self):
        with self.serial as port:
            return json.loads(port.read_until())
    def write(self, string):
        with self.serial as port:
            return port.write(str.encode(string+'\r\n'))
    def query(self, string):
        with self.serial as port:
            port.write(str.encode(string+'\r\n'))
            reply = port.read_until()
            if reply == b'':
                return None
            else:   
                return json.loads(reply)

    def reset(self):
        return self.query('RESET')
    def enable(self):
        return self.query('ENABLE')
    def disable(self):
        return self.query('DISABLE')
    #-----------------------------------#
    # Gets lockin current reading. Will try 5 times to read the UART mux line. If it fails, the lockin may require resetting if there is no no hardware Errors. Also Check for 5V on UART Mux (JST 2 Pin Connector)
    # Command is 'read param param'
    def mux_check(self, mux_true_command, mux_false_command):
        response = None
        if self.mux == True: 
            response = self.query(mux_true_command)
        else:
            response = self.query(mux_false_command)
        return response
    def temp_conversion(self, input_data):
        if type(input_data) == dict:
            return input_data
        if type(self.calibration_array) != None:
            # Expecting a message in the followin format: [32925, 1738841083, 2043.482]
            return round(np.interp(input_data, self.calibration_array[:,1], self.calibration_array[:,0]), 5)
        else:
            return input_data

    def read_temp(self):
        # mux_address cam be mux # or name of sensor
        command = 'R?'
        response = self.mux_check(mux_true_command= f'{command} {self.name}', mux_false_command = f'{command}')
        return self.temp_conversion(response)
    
    #-----------------------------------#
    def read_last_temp(self):
        command = 'LAST?'
        response = self.mux_check(mux_true_command= f'{command} {self.name}', mux_false_command = f'{command}')
        return self.temp_conversion(response)
    
    def read_running_average_temp(self):
        command = 'RA?'
        response = self.mux_check(mux_true_command= f'{command} {self.name}', mux_false_command = f'{command}')
        return self.temp_conversion(response)
    
    def set_bias(self, bias_int):
        command = 'BIAS'
        response = self.mux_check(mux_true_command= f'{command} {self.name} {bias_int}', mux_false_command = f'{command} {bias_int}')
        return response     
    
    def read_bias(self):
        command = 'BIAS?'
        response = self.mux_check(mux_true_command= f'{command} {self.name}', mux_false_command = f'{command}')
        return response  
    
    def set_rate(self):
        command = 'RATE'
        response = self.mux_check(mux_true_command= f'{command} {self.name}', mux_false_command = f'{command}')
        return response  
class Diode(object):
    def __init__(self, address, calibration_file = None, name = None):
        try:
            self.serial = Serial(address, timeout=5)
            self.serial.close()
        except Exception as e:
            print(e)
            pass
        if calibration_file != None:
            self.calibration_array = np.loadtxt(calibration_file, delimiter =',')
        else:
            self.calibration_array = None
        self.name = name
    def read(self):
        with self.serial as port:
            return json.loads(port.read_until())
    def write(self, string):
        with self.serial as port:
            return port.write(str.encode(string+'\r\n'))
    def query(self, string):
        with self.serial as port:
            port.write(str.encode(string+'\n'))
            reply = port.read_until()
            if reply == b'':
                return None
            else:   
                return json.loads(reply)
    def name_check(self, name_true_command, name_false_command):
        if type(self.name) != None: 
            response = self.query(name_true_command)
        elif type(self.name) == None:
            response = self.query(name_false_command)
        return response    
    
    def read_temp(self):
        # mux_address cam be mux # or name of sensor
        command = 'V?'
        response =  self.name_check(name_true_command = f'{command} {self.name}' , name_false_command = f'{command}')
        # Comment out dict reply. Should just reply a voltage now
        if response == None:
            return None
        else:
            for key in response:
                value = response[key]
                if type(value) == None:
                    response[key] = None
                else:
                    response[key] = self.temp_conversion(value)

            #return response
            return response[key]
    def temp_conversion(self, input_data):
        if  type(self.calibration_array) == None:
            return input_data
        else:
            return np.interp(input_data, self.calibration_array[:,1], self.calibration_array[:,0])
class PumpHeater(object):
    def __init__(self, address, name = None):
        try:
            self.serial = Serial(address,timeout=5)
            self.serial.close()
        except Exception as e:
            print(e)
            pass
        self.name = name
        self.been_set_enabled = False
    def read(self):
        with self.serial as port:
            return json.loads(port.read_until())
    def write(self, string):
        with self.serial as port:
            return port.write(str.encode(string+'\r\n'))
    def query(self, string):
        with self.serial as port:
            port.write(str.encode(string+'\n'))
            reply = port.read_until()
            if reply == b'':
                return None
            else:   
                return json.loads(reply)
            
    def name_check(self, name_true_command, name_false_command):
        if type(self.name) != None: 
            response = self.query(name_true_command)
        elif type(self.name) == None:
            response = self.query(name_false_command)
        return response    
    
    # Channel can be an int or a string
    #-----------------------------------#
    # Gets High power heater set dac value
    # Command is 'HPH_I? param'
    def get_pump_current_int(self):
         command = 'HPH_I?'
         return self.name_check(name_true_command = f'{command} {self.name}' , name_false_command = f'{command}')
    #-----------------------------------#
    # Gets High power heater measured voltage and current
    # Command is 'HPH_MON? param'
    def get_pump_measurement(self):
        command = 'HPH_MON?'
        reply = self.name_check(name_true_command = f'{command} {self.name}' , name_false_command = f'{command}')
        return reply[self.name]
    #-----------------------------------#
    # Gets High power heater enable/disable state
    # Command is 'HPH_EN? param'
    def get_enable(self):
        command = 'HPH_EN?'
        return self.name_check(name_true_command = f'{command} {self.name}' , name_false_command = f'{command}')
    #-----------------------------------#   
    # Sets High power heater dac value
    # Command is 'HPH_I param param'
    def set_pump_current(self, current):
        if type(current) == str:
            current = float(current)
        command = 'HPH_I'
        current = int(round(current, 0))
        print(current)
        self.name_check(name_true_command = f'{command} {self.name} {current}' , name_false_command = f'{command} {current}')
        if current != 0:
            if not self.been_set_enabled:
                print('enabled pump')
                self.set_pump_enable(1)
                self.been_set_enabled = True
        elif current == 0:
            self.set_pump_enable(0)
            self.been_set_enabled = False
    #-----------------------------------#   
    # Sets High power heater enable/disable state
    # Command is 'HPH_EN param param'
    def set_pump_enable(self, enabled):
        command = 'HPH_EN'
        return self.name_check(name_true_command = f'{command} {self.name} {enabled}' , name_false_command = f'{command} {enabled}')

class HeatSwitch(object):
    def __init__(self, address,name):
        try:
            self.serial = Serial(address)
            self.serial.close()
        except Exception as e:
            print(e)
            pass
        self.name = name
    def read(self):
        with self.serial as port:
            return json.loads(port.read_until())
    def write(self, string):
        with self.serial as port:
            return port.write(str.encode(string+'\r\n'))
    def query(self, string):
        with self.serial as port:
            port.write(str.encode(string+'\n'))
            reply = port.read_until()

            if reply == b'':
                return None
            else:   
                return json.loads(reply)
        
    # Channel can be an int or a string
    #-----------------------------------#
    # Gets Low Power Heater Voltage in mV
    # Commands is 'LPH_I? param'
    def get_heat_switch_voltage(self):
        reply = self.query(f'LPH_I? {self.name}')
        return reply[self.name]/1e3
    #-----------------------------------#
    # Sets Low Power Heater voltage Outlpg_put in mV
    # Command is 'LPH_I param param'
    # Forcing enable or disable output based on voltage value!
    def set_heat_switch_voltage(self, voltage = 0):
        command = 'LPH_I'
        # function require mV so converting V to mV
        if type(voltage) == str:
            voltage = float(voltage)
        voltage = voltage*1e3
        if voltage != 0:
            self.query(f'{command} {self.name} {voltage}')
            self.set_heat_switch_enable(1)
        elif voltage == 0:
            self.query(f'{command} {self.name} 0')
            self.set_heat_switch_enable(0)
    #-----------------------------------#
    # Gets Low power Heater Channel enable/disable state
    # Command is 'LPH_EN? param'
    def get_heat_switch_enable(self):
        return self.query(f'LPH_EN? {self.name}')
    #-----------------------------------#
    # Sets Low Power Heater enable/disable state
    # Command is 'LPH_EN param param'
    def set_heat_switch_enable(self, enabled):
        return self.query(f'LPH_EN {self.name} {enabled}')
    
    #-----------------------------------#

class WarmupHeater(object):
    def __init__(self, address, calibration_file = None, name = None):
        try:
            self.serial = Serial(address)
            self.serial.close()

        except Exception as e:
            print(e)
            pass
        self.name = name
    def read(self):
        with self.serial as port:
            return json.loads(port.read_until())
    def write(self, string):
        with self.serial as port:
            return port.write(str.encode(string+'/r/n'))
    def query(self, string):
        with self.serial as port:
            port.write(str.encode(string+'/r/n'))
            return json.loads(port.read_until())
        # Channel can be an int or a string
    #-----------------------------------#
    # Gets current measurement of warmup heater in mA
    # Command is 'I? param'
    def get_current(self, channel = 1):
            return self.query(f'I? {channel}')
    #-----------------------------------#
    # Sets ctrl pwm % decimal, see current_list files for what current to expect
    # Command is 'I param param'
    def set_current(self, current, channel = 1):
        return self.query(f'I {channel} {current}')
    #-----------------------------------#
    # Gets CTRL PWM value
    # Command is 'I_PWM? param'
    def get_current_pwm(self, channel = 1):
        return self.query(f'I_PWM? {channel}')
    #-----------------------------------#
    # Sets CTRL PWM value
    # Command is 'I_PWM param param'
    def set_current_pwm(self, pwm, channel = 1):
        return self.query(f'I_PWM {channel} {pwm}')
    #-----------------------------------#
    # Gets PWM value
    # Command is 'PWM? param'
    def get_pwm(self, channel = 1):
        return self.query(f'PWM? {channel}')
    #-----------------------------------#
    # Sets PWM value
    # Command is 'PWM param param'
    def set_pwm(self, pwm, channel = 1):
        return self.query(f'PWM {channel} {pwm}')
    #-----------------------------------#
    # Gets enable/disable state
    # Command is 'EN? param'
    def get_enable(self, channel = 1):
        return self.query(f'EN? {channel}')
    #-----------------------------------#
    # Sets enable/disable state
    # Command is 'EN param param'
    def set_enable(self, enabled, channel = 1):
        return self.query(f'EN {channel} {enabled}')
    def configure(self, current = None, pwm = None, enabled = None):
        current = None
        pwm = None
        enabled = None
        if current != None:
            current = self.set_current(current)
        if pwm != None:
            pwm = self.set_pwm(pwm)
        if enabled != None:
            enabled = self.set_enable(enabled)
        return current, pwm, enabled
    




# %%



#%%
from transitions import Machine
import random


# T4Ksetpt: 4.2
# T1Ksetpt: 2.8
# Tpump: 50
# T1Kfinal: 0.76
# Tpumpfinal: 6.7
# Tcheck: 0.76             #if the 1K sensor is below this temp the fridge does not recycle
# T1Khigh: 2
# Tswitch: 11

heaters = {
    "pump_heater": "Heater1",
}

criteria = {
    "pump_warming" :
        {
        "pump": 50,
        "1k": 0.1,
        },
    "heat_switch_cooling" :
        {
        "pump": 50,
        "switch": 11,
        },
}


class DummyHALClient:
    def send(self, message_dict):
        if message_dict['cmd'] == "get_temperatures":
            data = {
                "4k": 0.0,
                "1k": 0.0,
                "40k": 0.0,
            }
            return data
        elif message_dict['set_heater_value']:
            print(f"Setting heater {message_dict['name']} to {message_dict['value']}")

client = DummyHALClient()

class FridgeController1K:
    def __init__(self):
        pass

    def enable(self, heater):
        print("Enabling heater: ", heater)
        # client.send(

    def check_temperature_criteria(self):
        temperatures = client.send({"cmd": "get_temperatures"})
        for name, temp in criteria.items():
            if temperatures[name] > temp:
                return True
        temperatures['1k'] > criteria['1k']
    
    def on_enter_pump_warming(self):
        self.enable['pump_heater']
        print("Entered pump_warming state")

    # def advance_to_next_state(self):
    #     print("Advancing to next state")

#
# Initialize the state machine
fridge = FridgeController1K()
states = ['warm', 'heat_switch_cooling', 'pump_warming', 'cold', 'warming_up']
machine = Machine(model=fridge, states=states, initial='heat_switch_cooling')

# Advance from heat_switch_cooling to pump_warming only if we've satisfied the condition
# that we've reached our temperature criteria of interest.  
machine.add_transition(trigger='advance_to_next_state', source='heat_switch_cooling', dest='pump_warming',
                        conditions = 'check_temperature_criteria')
 # Internal transition; no state change; will not execute on_enter_heat_switch_cooling
machine.add_transition(trigger='advance_to_next_state', source='heat_switch_cooling', dest=None)


machine.add_transition(trigger='advance_to_next_state', source='pump_warming', dest='cooling',
                        conditions = 'check_temperature_criteria')
 # Internal transition; no state change; will not execute on_enter_heat_switch_cooling
machine.add_transition(trigger='advance_to_next_state', source='pump_warming', dest=None)



while True:
    fridge.advance_to_next_state()
    fridge.update_pid()
    if message is 'recycle':
        fridge.go_to_state('heat_switch_cooling')


#%%



# class FridgeController:

#     # Define some states. Most of the time, narcoleptic superheroes are just like
#     # everyone else. Except for...
#     

#     def __init__(self):

    #     # Superheroes need to keep in shape.
    #     machine.add_transition('work_out', 'hanging out', 'hungry')

    #     # Those calories won't replenish themselves!
    #     self.machine.add_transition('eat', 'hungry', 'hanging out')

    #     # Superheroes are always on call. ALWAYS. But they're not always
    #     # dressed in work-appropriate clothing.
    #     self.machine.add_transition('distress_call', '*', 'saving the world',
    #                      before='change_into_super_secret_costume')

    #     # When they get off work, they're all sweaty and disgusting. But before
    #     # they do anything else, they have to meticulously log their latest
    #     # escapades. Because the legal department says so.
    #     self.machine.add_transition('complete_mission', 'saving the world', 'sweaty',
    #                      after='update_journal')

    #     # Sweat is a disorder that can be remedied with water.
    #     # Unless you've had a particularly long day, in which case... bed time!
    #     self.machine.add_transition('clean_up', 'sweaty', 'asleep', conditions=['is_exhausted'])
    #     self.machine.add_transition('clean_up', 'sweaty', 'hanging out')

    #     # Our NarcolepticSuperhero can fall asleep at pretty much any time.
    #     self.machine.add_transition('nap', '*', 'asleep')

    # def update_journal(self):
    #     """ Dear Diary, today I saved Mr. Whiskers. Again. """
    #     self.kittens_rescued += 1

    # @property
    # def is_exhausted(self):
    #     """ Basically a coin toss. """
    #     return random.random() < 0.5

    # def change_into_super_secret_costume(self):
    #     print("Beauty, eh?")


import numpy as np
import zmqhelper as zmqh
from extractor_class import TrevisanExtractorRun
import json as json
import time
import PEF_Calculator as PEF
import data_loading_mod as dlm 
import base64

class ExtractorServer(zmqh.Server):
    def __init__(self, port, n_workers, aggregate=True):
        if aggregate:
            self.dataFormat = [('SA','u1'),('SB','u1'),('OA','u1'),('OB','u1')]
        else:
            self.dataFormat = [('SA','u1'),('SB','u1'),('OA','u8'),('OB','u8')]

        print(self.dataFormat)
        super().__init__(port, n_workers)

    # def handle(self, msg):
    #     '''
    #     To save time, one could split this up.
    #     Just send over the outcomes and seed first, then run
    #     extractor_object.write_input() and extractor_object.write_seed().
    #     Once entropy is send, use extractor_object.get_entropy(entropy) to
    #     update the entropy and run with extractor_object.execute_extractor()

    #     Inputs: A dictionary serialized with json containing the experimental
    #     outcomes, the seed (experiment settings), and the entropy.

    #     Output: a string of '1's and '0's encoded into a python bytes object.
    #     '''
    #     # Turn the bytes recieved into a string
        

    def run_extractor(self,params):
        print('')
        binData = self.convert_str_to_bytes(params['data'])
        params['data'] = None
        data = dlm.read_data_buffer(binData, self.dataFormat)
        binData = None

        outcomesReordered = np.array([[data['OA']],[data['OB']]])
        data = None
        outcomesReordered = outcomesReordered.transpose().flatten()
        # print('step 1')
        # print('')

        nTrials = int(params['stoppingCriteria'])
        if nTrials>-1:
            nBits = 2*nTrials
            # print('nBits', nBits, len(outcomesReordered))
            if len(outcomesReordered)>nBits:
                print('data too long, truncate to stoppingCriteria')
                outcomesReordered = outcomesReordered[0:nBits]
            else:
                # Need to pad out the results
                print('data too short, pad to stoppingCriteria')
                outcomesPadded = np.zeros(nBits)
                outcomesPadded[0:len(outcomesReordered)] = outcomesReordered
                outcomesReordered = outcomesPadded

        # print('step 2')
        outcomesReordered = outcomesReordered.astype(int)

        # print('get rid of vals >1')
        # outcomes[outcomes>0] = 1
        # print('before to list', outcomesReordered[0:100])
        # print('to list')
        outcomesReordered = outcomesReordered.tolist()
        print('OUTCOMES', outcomesReordered[0:100])

        seed = np.array(params['seed']).tolist()
        entropy = params['entropy']
        nBitsOut = int(params['nBitsOut'])
        errorExtractor = float(params['errorExtractor'])

        # extractorObject = TrevisanExtractorRun(params['outcomesReordered'], params['seed'], params['entropy'], params['nbits'], params['error_prob'])
        extractorObject = TrevisanExtractorRun(outcomesReordered, seed, entropy, nBitsOut, errorExtractor)
        # Write the input and seed
        print('extractor object created')
        extractorObject.write_input()
        print('write input')
        extractorObject.write_seed()
        print('write seed')
        extractorObject.execute_extractor()
        print('reading output')
        outBits = extractorObject.read_output()
        print('output bits', outBits)
        print('cleaning up')
        extractorObject.remove_files() 
        extractorObject = None
        print('files deleted, ready for more input')
        print('')

        return outBits#.encode('utf-8')

    def get_delta(self, isQuantum):
        if isQuantum:
            delta = 4E-8
        else:
            delta = 0
        return delta

    def get_freqs(self, params):
        binData = self.convert_str_to_bytes(params['data'])
        # binData = binData.tobytes()
        data = dlm.read_data_buffer(binData, self.dataFormat)
        # Truncate data to the stopping criteria
        # if params['stoppingCriteria']:
        if 'stoppingCriteria' in params:
            stoppingCriteria = int(params['stoppingCriteria'])
        else:
            stoppingCriteria = -1 
        data = data[0:stoppingCriteria]
        freq = dlm.get_freqs(data)
        freq = freq
        # print(freq, type(freq))
        return freq

    def extract_data(self, data):
        pass 

    def find_optimal_beta(self, params):
        freq = np.array(params['freq'])
        epsilonBias = float(params['epsilonBias'])
        # delta = float(params['delta'])
        nBitsOut = int(params['nBitsOut'])
        # error = float(params['error'])
        errorSmoothness = params['errorSmoothness']
        errorExtractor = params['errorExtractor']
        # fracSmoothness = float(params['fracSmoothness'])
        isQuantum = bool(params['isQuantum'])
        delta = self.get_delta(isQuantum)

        beta = PEF.find_optimal_beta(freq, epsilonBias, delta, nBitsOut, errorSmoothness, errorExtractor, isQuantum)
        return beta

    def calc_PEFs(self, params):
        freq = np.array(params['freq'])
        beta = float(params['beta'])
        epsilonBias = float(params['epsilonBias'])
        isQuantum = bool(params['isQuantum'])
        delta = self.get_delta(isQuantum)
        pefs, gain = PEF.calc_PEFs(freq, beta, epsilonBias, delta)
        print(pefs)
        print(gain)
        return pefs, gain

    def calc_entropy(self, params):
        freq = np.array(params['freq'])
        pefs = np.array(params['pefs'])
        beta = float(params['beta'])
        epsilonBias = float(params['epsilonBias'])
        errorSmoothness = float(params['errorSmoothness'])
        isQuantum = bool(params['isQuantum'])
        delta = self.get_delta(isQuantum)

        entropy = PEF.calculate_entropy(freq, pefs, errorSmoothness, 
                beta, epsilonBias, delta, isQuantum=isQuantum)
        return entropy

    def compute_extractor_properties(self, params):
        nBitsOut = int(params['nBitsOut'])
        gain  = float(params['gain'])
        errorSmoothness = float(params['errorSmoothness'])
        errorExtractor = float(params['errorExtractor'])
        beta  = float(params['beta'])
        gain  = float(params['gain'])
        epsilonBias  = float(params['epsilonBias'])
        isQuantum = bool(params['isQuantum'])

        nBitsThreshold = PEF.calc_threshold_bits(nBitsOut, errorExtractor, isQuantum=isQuantum)
        nTrialsNeeded = PEF.compute_minimum_trials(nBitsOut, beta, gain, errorSmoothness, isQuantum=isQuantum)
        nBitsIn = 2*nTrialsNeeded
        seedLength = PEF.calc_seed_length(nBitsOut, nBitsIn, errorExtractor, isQuantum=isQuantum)

        return nBitsThreshold, nTrialsNeeded, seedLength

    def process_entropy(self, params):
        # nBitsThreshold, nTrialsNeeded, seedLength = self.compute_extractor_properties(params)
        # params['nBitsThreshold'] = nBitsThreshold 
        freq = self.get_freqs(params) 
        params['freq'] = freq
        entropy = self.calc_entropy(params)

        nBitsThreshold = float(params['nBitsThreshold'])
        success = bool(entropy>nBitsThreshold)
        return entropy, success

    def get_experiment_parameters(self, params):
        # params['stoppingCriteria'] = -1
        # freq = self.get_freqs(params) 
        # params['freq'] = freq
        print('finding optimal_beta')
        beta = self.find_optimal_beta(params)
        params['beta'] = beta
        print('beta', beta)

        pefs, gain = self.calc_PEFs(params)
        params['pefs'] = pefs 
        params['gain'] = gain
        print('pefs', pefs)
        print('gain', gain)

        nBitsThreshold, nTrialsNeeded, seedLength = self.compute_extractor_properties(params)
        return pefs, beta, gain, nBitsThreshold, nTrialsNeeded, seedLength

    def handle(self,msg):
        print('')
        print('message received')
        inputs = json.loads(msg)
        cmd = inputs['cmd']
        params = inputs['params']
        # print("Received request: %s" % cmd)
        # print('cmd', cmd[0])
        cmd = cmd.lower()
        print('Received command:', cmd)

        # Msgout is returned from motor command
        try:
            if cmd == "extract":
                outBits = self.run_extractor(params)
                res = {}
                res['outBits'] = outBits

            elif cmd == 'freqs':
                freqs = self.get_freqs(params)
                res = {}
                res['freqs'] = freqs
                # msgout = freqs
                
            elif cmd == 'calc_pefs':
                pefs, gain = self.calc_PEFs(params)
                res = {}
                res['pefs'] = pefs
                res['gain'] = gain

            elif cmd == 'find_beta':
                beta = self.find_optimal_beta(params)
                res = {}
                res['beta'] = beta
                
            elif cmd == 'calc_entropy':
                entropy = self.calc_entropy(params)
                res = {}
                res['entropy'] = entropy
                
            elif cmd == 'process_entropy':
                entropy, success = self.process_entropy(params)
                res = {}
                res['entropy'] = entropy
                # res['nBitsThreshold'] = nBitsThreshold
                res['isThereEnoughEntropy'] = success
                print('entropy', entropy, success)

            elif cmd == 'calc_extractor_properties':
                nBitsThreshold, nTrialsNeeded, seedLength = self.compute_extractor_properties(params)
                res = {}
                res['nBitsThreshold'] = nBitsThreshold
                res['nTrialsNeeded'] = nTrialsNeeded
                res['seedLength'] = seedLength
                
            elif cmd == 'get_experiment_parameters':
                pefs, beta, gain, nBitsThreshold, nTrialsNeeded, seedLength = self.get_experiment_parameters(params)
                res = {}
                res['pefs'] = pefs
                res['beta'] = beta
                res['gain'] = gain
                res['nBitsThreshold'] = nBitsThreshold
                res['nTrialsNeeded'] = nTrialsNeeded
                res['seedLength'] = seedLength

            else:
                res = {}
                res['error'] = "Invalid Command"

        # Catch errors and return them
        except Exception as e:
            print("Error: %r" % e)
            res = {}
            res['error'] = "Error: "+str(e)
            raise e
        msgout = self.encode_message_to_JSON(res)
        msgout = msgout.encode('utf-8')

        return msgout

    def convert_str_to_bytes(self, strData):
        data = base64.b64decode(strData)
        return data

    def convert_bytes_to_str(self, binData):
        strData = base64.b64encode(binData).decode('utf-8')
        return strData

    def encode_message_to_JSON(self, result):
        for key, value in result.items():
            if type(value)==bytes:
                value = convert_bytes_to_str(value)
            if type(value)==np.ndarray:
                value = value.tolist()
            # if isinstance(value, bool):
            #     return str(value).lower()
            result[key] = value
        msgJSON = json.dumps(result)
        return msgJSON


if __name__ == '__main__':
    print('Starting Extractor Server')
    pefs = ExtractorServer(port='5553', n_workers=1, aggregate=True)


import requests
import json

class MonitorClient:
    def __init__(self, url, timeout = 0.1):
        self.url = url
        self.timeout = timeout

    def get_metrics(self, name = None):
        """ Get the metrics from the monitor server as a dictionary.
        Optionally, specify which subset of metrics (e.g. temperatures,
        heater_max_values) by specifying `name """
        response = requests.get(self.url, timeout = self.timeout)
        metrics_dict = json.loads(response.text)
        if name is not None:
            return metrics_dict[name]
        else:
            return metrics_dict


#%%
import concurrent.futures
import requests
import json
import pandas as pd
import psycopg2
import tomllib
import datetime
import time

class Scraper:
    """ A multi-threaded HTTP scraper that fetches JSON data from a list of URLs. """
    def __init__(self,  timeout = 1, num_workers = 10):
        self.timeout = timeout
        self.num_workers = num_workers

    def fetch(self, url):
        try:
            response = requests.get(url, timeout = self.timeout)
            return (url, response.status_code, response.text)
        except requests.RequestException as e:
            return (url, None, str(e))

    def scrape(self, urls):
        responses = {}
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.num_workers) as executor:
            future_to_url = {executor.submit(self.fetch, url): url for url in urls}
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
                try:
                    url, status_code, content = future.result()
                    if status_code == 200:
                        responses[url] = json.loads(content)
                except Exception as exc:
                    print(f"URL: {url} generated an exception: {exc}")
        return responses


class PostgresUploader:

    def __init__(self, host, port, user, password, database, timeout = 1):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.timeout = timeout

    def scraped_responses_to_df(self, responses):
        """
        Convert a dictionary of responses into three separate pandas DataFrames.
        Parameters:
        responses (dict): A dictionary containing response data. Each key is a unique identifier, 
                            and each value is a dictionary
        Returns:
        tuple: A tuple containing three pandas DataFrames:
                - df_temperatures: DataFrame with columns ['time', 'cryostatname', 'sensorname', 'temperature']
                - df_heaters: DataFrame with columns ['time', 'cryostatname', 'heatername', 'value']
                - df_state: DataFrame with columns ['time', 'cryostatname', 'state']
        """

        temperatures_df_list = []
        heaters_df_list = []
        state_df_list = []
        current_time = datetime.datetime.now(datetime.timezone.utc)
        for response in responses.values():
            cryostat_name = response['metadata']['cryostat_name']
            if 'temperatures' in response:
                df = pd.DataFrame({
                    'time' : current_time,
                    'name' :  cryostat_name,
                    'sensor': response['temperatures'].keys(),
                    'value':  response['temperatures'].values()}
                    )
                temperatures_df_list.append(df)
            if 'heaters' in response:
                df = pd.DataFrame({
                    'time' : current_time,
                    'cryostat' :  cryostat_name,
                    'name': response['heaters'].keys(),
                    'value':  response['heaters'].values()}
                    )
                heaters_df_list.append(df)
            if 'state' in response:
                df = pd.DataFrame({
                    'time' : current_time,
                    'cryostat' :  [cryostat_name],
                    'state': [response['state']]},
                    )
                state_df_list.append(df)
        if len(temperatures_df_list) > 0:
            df_temperatures = pd.concat(temperatures_df_list)
        else:
            df_temperatures = None
        if len(heaters_df_list) > 0:
            df_heaters = pd.concat(heaters_df_list)
        else:
            df_heaters = None
        if len(state_df_list) > 0:
            df_state = pd.concat(state_df_list)
        else:
            df_state = None

        return df_temperatures, df_heaters, df_state


    def upload_dataframe_to_table(self, df, table_name):
        """
        Uploads the entries of a pandas DataFrame to a table in a PostgreSQL database.

        Parameters:
        df (pd.DataFrame): The DataFrame to upload.
        table_name (str): The name of the table to upload the data to.
        """
        # Establish a connection to the database
        with psycopg2.connect(
                        host = self.host,
                        port = self.port,
                        user = self.user,
                        password = self.password,
                        dbname = self.database,
                        connect_timeout = self.timeout) as conn:
            with conn.cursor() as cursor:
                # Iterate over DataFrame rows
                for index, row in df.iterrows():
                    # Construct the SQL INSERT statement
                    columns = ', '.join(row.index)
                    values = ', '.join([f'%({col})s' for col in row.index])
                    sql = f"INSERT INTO {table_name} ({columns}) VALUES ({values})"
                    
                    # Execute the SQL command
                    cursor.execute(sql, row.to_dict())
                
                # Commit the transaction
                conn.commit()




import fridgeos.zmqhelper as zmqhelper
import time
import json


class HALClient:
    def __init__(self, ip, port):
        self.connection = zmqhelper.Client(ip, port)

    def send_command(self, command, **kwargs):
        command_json = {"cmd": command}
        command_json.update(kwargs)
        message = json.dumps(command_json)
        response = self.connection.send_message(message)
        return json.loads(response)
    
    def get_temperatures(self):
        return self.send_command("get_temperatures")
    
    def get_temperature(self, name):
        return self.send_command("get_temperature", name = name)
    
    def set_heater_value(self, name, value):
        return self.send_command("set_heater_value", name = name, value = value)
    
    def get_heater_values(self):
        return self.send_command("get_heater_values")
    
    def get_heater_max_values(self):
        return self.send_command("get_heater_max_values")
    
if __name__ == "__main__":
    halclient = HALClient(ip = '127.0.0.1', port = '5555')
    halclient.get_temperatures()



import zmq
import threading


class Server():
    def __init__(self, port, n_workers=4):
        self.port = str(port)
        url_worker = "inproc://workers"
        url_client = 'tcp://*:'+self.port
        self.context = zmq.Context.instance()
        # Socket to talk to clients
        clients = self.context.socket(zmq.ROUTER)
        clients.bind(url_client)

        workers = self.context.socket(zmq.DEALER)
        workers.bind(url_worker)

        self.workers = []
        for k in range(0, n_workers):
            self.workers.append(threading.Thread(target=self.start_worker, daemon=True, args=(url_worker,k,)))
            self.workers[-1].start()

        zmq.proxy(clients, workers)
        # We never get here but clean up anyhow
        clients.close()
        workers.close()
        self.context.term()


    def start_worker(self, url_worker, k=0):
        """Worker routine"""
        context = self.context or zmq.Context.instance()
        # Socket to talk to dispatcher
        socket = context.socket(zmq.REP)
        socket.connect(url_worker)

        while True:
            message = socket.recv()
            message = message.decode()
            response = self.handle(message)
            # print('worker '+str(k)+' reporting for duty, '+response.decode())
            socket.send_string(response)

    def handle(self, message):
        return message.encode()

if __name__ == '__main__':
    zmqs = Server('5555', n_workers=4)



#   Author: Krister Shalm
#   Simple ZMQ client in Python
#   Connects REQ socket

import zmq
import time
import json
class Client():
    """

    """
    def __init__(self, ip, port):
        self.ip = 'tcp://'+str(ip)
        self.port = str(port)
        self.simple_init()

    def simple_init(self):
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.REQ)
        self.socket.connect(self.ip + ":" + self.port)

        self.poller = zmq.Poller()
        self.poller.register(self.socket, zmq.POLLIN)
        self.connected = True


    def send_message(self, msg, timeout=3):
        msg = json.dumps(msg)
        msgTimeout = 'Timeout'
        # msgTimeout = msgTimeout.encode()
        
        try:
            self.socket.send_string(msg)
        except Exception as e:
            print(e)
            return(msgTimeout)
        socks = dict(self.poller.poll(timeout))
        if socks:
            if socks.get(self.socket) == zmq.POLLIN:
                try:
                    response = self.socket.recv()
                    # print(msg)
                    # print(response)
                    # print('')
                    response = response.decode()
                except:
                    # response = msgTimeout
                    self.connected = False
                    self.reconnect()
                return response
        else:
            response = msgTimeout
            self.connected = False
            self.reconnect()


        return(response)

    def reconnect(self):
        self.close()
        self.simple_init()


    def close(self):
        self.socket.close()
        self.context.term()


if __name__ == '__main__':
    import time
    ip = 'qittlab-nuc-05.campus.nist.gov'
    port = '5555'
    print('creating connection')
    con = Client(ip, port)
    clientID = time.time()
    '''
    for i in range(100):
        resp = con.send_message('clientID:'+str(clientID)+' '+str(i))
        time.sleep(0.1)
        print(resp)
    '''



import zmqhelper as zmqh
import json

class MyCustomServer(zmqh.Server):

    # Overload the handle function here
    def handle(self,msg):
        print('')
        print('message received')
        inputs = json.loads(msg)
        cmd = inputs['cmd']
        params = inputs['params']
        # print("Received request: %s" % cmd)
        # print('cmd', cmd[0])
        cmd = cmd.lower()
        print('Received command:', cmd)

        # Msgout is returned from motor command
        try:
            if cmd == "extract":
                outBits = self.run_extractor(params)
                res = {}
                res['outBits'] = outBits

            elif cmd == 'freqs':
                freqs = self.get_freqs(params)
                res = {}
                res['freqs'] = freqs
                # msgout = freqs
                
            elif cmd == 'calc_pefs':
                pefs, gain = self.calc_PEFs(params)
                res = {}
                res['pefs'] = pefs
                res['gain'] = gain

            elif cmd == 'find_beta':
                beta = self.find_optimal_beta(params)
                res = {}
                res['beta'] = beta
                
            elif cmd == 'calc_entropy':
                entropy = self.calc_entropy(params)
                res = {}
                res['entropy'] = entropy
                
            elif cmd == 'process_entropy':
                entropy, success = self.process_entropy(params)
                res = {}
                res['entropy'] = entropy
                # res['nBitsThreshold'] = nBitsThreshold
                res['isThereEnoughEntropy'] = success
                print('entropy', entropy, success)

            elif cmd == 'calc_extractor_properties':
                nBitsThreshold, nTrialsNeeded, seedLength = self.compute_extractor_properties(params)
                res = {}
                res['nBitsThreshold'] = nBitsThreshold
                res['nTrialsNeeded'] = nTrialsNeeded
                res['seedLength'] = seedLength
                
            elif cmd == 'get_experiment_parameters':
                pefs, beta, gain, nBitsThreshold, nTrialsNeeded, seedLength = self.get_experiment_parameters(params)
                res = {}
                res['pefs'] = pefs
                res['beta'] = beta
                res['gain'] = gain
                res['nBitsThreshold'] = nBitsThreshold
                res['nTrialsNeeded'] = nTrialsNeeded
                res['seedLength'] = seedLength

            else:
                res = {}
                res['error'] = "Invalid Command"

        # Catch errors and return them
        except Exception as e:
            print("Error: %r" % e)
            res = {}
            res['error'] = "Error: "+str(e)
            raise e
        msgout = self.encode_message_to_JSON(res)
        msgout = msgout.encode('utf-8')

        return msgout


#%%
import json
import tomllib
import time
import numpy as np
import logging
import sys
import fridgeos.zmqhelper as zmqhelper
import os

# Device driver imports
from .drivers.haldrivers import hal_classes

# TODO: Make 1 worker per communication address (e.g. 1 for COM5, 1 for COM6, 1 for /dev/usb321)
# TODO: Add configuration-file checking (e.g. for max_heater_value) and error reporting

class HALServer(zmqhelper.Server):

    def __init__(self, port, hardware_toml_path,
                 debug = False, n_workers = 1,
                 log_path = None):
        self.hardware = {}
        self.hardware['thermometers'] = {}
        self.hardware['heaters'] = {}
        self.load_hardware(hardware_toml_path)
        self._setup_logging(log_path, debug)
        super().__init__(port, n_workers)
        print('HAL Server started')
    
    def get_hardware(self, name, hardware_type):
        """ Checks that a device with the given name exists
         then returns its python object for use """
        if name not in self.hardware[hardware_type]:
            self.logger.error(f'Device name "{name}" not found for hardware type {hardware_type}, available options are {self.hardware[hardware_type].keys()}')
            raise ValueError(f'Device name "{name}" not found for hardware type {hardware_type}')
        else:
            return self.hardware[hardware_type][name]['python_object']

    def _setup_logging(self, log_path, debug = False):
        # Make the logging path if doesn't exist
        if log_path is not None:
            log_dir = os.path.dirname(log_path)
            if not os.path.exists(log_dir):
                os.makedirs(log_dir)
        # Create the logger
        logger = logging.getLogger('HAL')
        logger.setLevel(logging.DEBUG)
        # Create 3 "handlers" two which output to a log file, the other to stdout
        handler1 = logging.StreamHandler(sys.stdout)
        handler2 = logging.FileHandler(os.path.join(log_path, 'hal-errors.log'), mode='a')
        handler3 = logging.FileHandler(os.path.join(log_path, 'hal-debug.log'), mode='a')
        handler1.setLevel(logging.INFO)
        handler2.setLevel(logging.INFO)
        handler3.setLevel(logging.DEBUG)
        # Set the format of the log messages
        log_date_format = '%Y-%m-%d %H:%M:%S'
        format = logging.Formatter(fmt = '%(asctime)s.%(msecs)03d,\t%(levelname)s,\t%(message)s', datefmt=log_date_format)
        handler1.setFormatter(format)
        handler2.setFormatter(format)
        handler3.setFormatter(format)
        # Create the logger
        logger.addHandler(handler1)
        if log_path is not None:
            logger.addHandler(handler2)
            if debug is True:
                logger.addHandler(handler3)
        self.logger = logger
        self.logger.debug('Starting up server')

    def handle(self, message):
        message_dict = json.loads(message)
        command = message_dict['cmd'].lower()
        self.logger.debug(f"Message received: '{message}'")
        
        try:
            if command == 'get_temperature':
                name = message_dict['name']
                output = self.get_temperature(name)
            elif command == 'get_temperatures':
                output = self.get_temperatures()
            elif command == 'set_heater_value':
                name = message_dict['name']
                value = message_dict['value']
                output = self.set_heater_value(name, value)
            elif command == 'get_heater_value':
                name = message_dict['name']
                output = self.get_heater_value(name)
            elif command == 'get_heater_values':
                output = self.get_heater_values()
            elif command == 'get_heater_max_values':
                output = self.get_heater_max_values()
            else:
                self.logger.error(f'Unrecognized command "{command}"')
        # Catch errors, log them, and return an empty dictionary
        except Exception as e:
            self.logger.error('Python error:', exc_info=e)
            output = {}

        message_out = json.dumps(output)
        self.logger.debug(f"Sending message: '{message_out}'")
        return message_out               

    def load_hardware(self, hardware_toml_path):
        # Load the TOML file
        with open(hardware_toml_path, "rb") as f:
            all_hardware = tomllib.load(f)
        # For each type of hardware (e.g heater/thermometer), go through each
        # device in the hardware_list, create a Python object, and set it up
        for hardware_type, hardware_list in all_hardware.items():
            # Check for duplicate names
            names = [h['name'] for h in hardware_list]
            if len(names) != len(set(names)):
                duplicates = {name for name in names if names.count(name) > 1}
                raise ValueError(f'Duplicate names found in configuration file, {hardware_type} section: {duplicates}')
            # Set up each individual device
            for hw in hardware_list:
                hw_name = hw['hardware']
                if hw_name not in hal_classes:
                    raise ValueError(f'Unrecognized/no driver for {hardware_type} named "{hw_name}"')
                # Create the device as a python object
                hal_class = hal_classes[hw_name]
                python_object = hal_class()
                print(f'Attempting to setup {hw_name} hardware with setup arguments {hw}["setup"]')
                # Configure the device
                print(hw)
                python_object.setup(**hw['setup'])
                print(f'Added {hw["hardware"]} thermometer successfully')
                # Add the thermometer object to self.thermometers dictionary
                hw['python_object'] = python_object
                name = hw.pop('name')
                self.hardware[hardware_type][name] = hw

    def get_temperature(self, name):
        """ Get the temperature of a single thermometer """
        hw = self.get_hardware(name = name, hardware_type = 'thermometers')
        return {name : hw.get_temperature()}
    
    def get_heater_value(self, name):
        """ Get the value of a single heater """
        hw = self.get_hardware(name = name, hardware_type = 'heaters')
        return {name : hw.get_heater_value()}
    
    def set_heater_value(self, name, value):
        """ Set the value of a single heater """
        hw = self.get_hardware(name = name, hardware_type = 'heaters')
        return {name : hw.set_heater_value(value)}

    def get_temperatures(self):
        temperatures = {}
        for name in self.hardware['thermometers'].keys():
            temperatures.update(self.get_temperature(name))
        return temperatures
    
    def get_heater_values(self):
        """ Get the values of all heaters, returns a dictionary of the
        form {name1 : value1, name2 : value2, ...} """
        values = {}
        for name in self.hardware['heaters'].keys():
            values.update(self.get_heater_value(name))
        return values
    
    def get_heater_max_values(self):
        values = {}
        for name in self.hardware['heaters'].keys():
            values[name] = self.hardware['heaters'][name]['max_value']
        return values
    

    # def find_unique_addresses(self,toml_file_path):
    #     # Load the TOML file
    #     data = tomllib.load(toml_file_path)
    #     # Initialize a set to store unique addresses
    #     unique_addresses = set()
    #     # Look for "address" keys in all thermometer, heater, and relay sections
    #     for section_name in list(data.keys()):
    #         if section_name in data:
    #             for item in data[section_name]:
    #                 if "address" in item:
    #                     unique_addresses.add(item["address"])
    #     return unique_addresses




# -*- coding: utf-8 -*-
"""
Created on Wed Sep 11 12:25:56 2024

@author: Ryan Morgenstern @ NIST
"""
#%%
from statemachine import State, StateMachine
import time
import tomllib

#from timer import ElapsedTime
from simple_pid import PID

class crc_cmd(StateMachine):
    def __init__(self, settings_toml, hal_client, monitor_client):
        # Add ZMQ connection - HAL Client
        self.hal_client = hal_client
        # Add html scrapper - Monitor Client
        self.monitor_client = monitor_client
        # Import settings from toml file
        with open(f'{settings_toml}', "rb") as f:
            self.sm_settings = tomllib.load(f)
        
        # Keep track of time when we enter states. It is set on_entry into new state
        self.enter_state_time = 0
        #Check if just entered state
        self.just_entered = False
        # Controls if the state machine is running (Must call start_state_machine() )
        self.cycle_running = False
        # Time.sleep before state is ran again
        self.state_machine_cycle_time = self.sm_settings['misc']['state_machine_cycle_time'] #seconds
        # Holds all fridge data from metric server. refreshed every state_machine_cycle_time
        self.fridge_state = {}
        self.fridge_timer = ElapsedTime()
        self.dt = 0
        # Varaible for when the still stays activated and not turned off for PID control
        self.still_activated = False
        # The current_cycle will be defined when entering eithe the start_cycle_A or start_cycle_B state. It's used to leave turn_on_still_heater state to the appropriate cycle
        self.current_cycle = None
        # PID controller setup
        self.pid = {}
        # Get max values of HAL TOML hardware by grabbing metric server data
        self.update_fridge_state()
        self.setup_pids()
        self.currently_in_use_pids = {}
        self.time_elapse_dict = {}
        super().__init__()
        # Allows states to move on without transition (required dont touch)
        self.allow_event_without_transition = 1

    def setup_pids(self):
        for key,values in self.sm_settings['pid_configuration'].items():
            max_value = self.fridge_state['heater_max_values'][key]
            self.pid[key] = PID(Kp = values['kp'], 
                                        Kd = values['kd'],
                                        Ki = values['ki'],
                                        sample_time=None,
                                        output_limits = (0, max_value))
            
    #-----------------------------------------------------------------------------------------------------#
    #                                    Defining states of state machine
    #-----------------------------------------------------------------------------------------------------#
    # Wait for the compressor to cool the the heat switches to below < 12K before starting state machine
    wait_for_heat_switches_to_cool = State(name = 'Waiting for heat switches to cool < 12K' , initial = True, value = 1)
    #-----------------------------------------------------------------------------------------------------#
    # Turn on all pumps and regulate to 45K to release He3 & He4 till head temperatures are < 4.1K
    turn_on_all_pumps = State(name = 'Cooling heads < 4K', value = 2)
    #-----------------------------------------------------------------------------------------------------#
    # Cool down the He4 Heads by turning off 4 pump A & B and turning on 4 switch A & B
    turn_on_4_heat_switches = State(name = 'Cooling He4 Heads A & B', value = 3)
    #-----------------------------------------------------------------------------------------------------#
    # Cooldown the He3 heads by turning off 3 pump A & B and turning on 3 switch A & B
    turn_on_3_heat_switches = State(name = 'Cooling He3 Heads A & B', value = 4)
    #-----------------------------------------------------------------------------------------------------#
    # Prepare cycle A by turning off 3 & 4 heat switch A.  Check that they are 0V / Dac value and let them cool < 12K.
    start_cycle_A = State(name = 'Starting Cycle A', value = 5)
    #-----------------------------------------------------------------------------------------------------#
    # I believe this state was added to not thermally shock the system when heating to 45K. 
    get_pumps_warm_A = State(name = 'Warming 3 & 4 A pumps', value = 6)
    #-----------------------------------------------------------------------------------------------------#
    # Turn on 3 & 4 pumps A for a period of time "30*60*1000" and if "cold" line 557 s_m.js
    get_pumps_hot_A = State(name = 'Getting 3 & 4 A pumps hot', value = 7)
    #-----------------------------------------------------------------------------------------------------#
    # Turn on heat switch 4A to cool down 4 head A, check if (3HeadA < 6 & 4HeadA < 1) or 40*60*1000 elapsedtime
    turn_on_heat_switch_4_A = State(name = 'Turn on heat switch 4 A', value = 8)
    #-----------------------------------------------------------------------------------------------------#
    # Turn off 3pumpA and turn on heat switch 3 A. Wait for elapse time
    turn_on_heat_switch_3_A = State(name = 'Turn on heat switch 3 A', value = 9)
    #-----------------------------------------------------------------------------------------------------#
    # Start cycle B by turning off 3 & 4 heat switch B and check that they are 0V / Dac value. Let them cool < 12K
    start_cycle_B = State(name = 'Starting Cycle B', value = 10)
    #-----------------------------------------------------------------------------------------------------#
    # I believe this state was added to not thermally shock the system when heating to 45K. 
    get_pumps_warm_B = State(name = 'Warming 3 & 4 B pumps', value = 11)
    #-----------------------------------------------------------------------------------------------------#
    # Turn on pumps 3 & 4 B to 45K for a period of time and untii 3 and 4 head times less than 6K & 4K
    get_pumps_hot_B = State(name = 'Getting 3 & 4 B pumps hot', value = 12)
    #-----------------------------------------------------------------------------------------------------#
    # Turn on heat switch 4 B to cool down 4 head B, check if (3HeadB < 6 & 4HeadB < 1) or 40*60*1000 elapsedtime
    turn_on_heat_switch_4_B = State(name ='Turn on heat switch 4 B', value = 13)
    #-----------------------------------------------------------------------------------------------------#
    # Turn off 3pumpB and turn on heat switch 3 B. Wait for elapse time
    turn_on_heat_switch_3_B = State(name = 'Turn on heat switch 3 B', value = 14)
    #-----------------------------------------------------------------------------------------------------#
    # Once cold still is below 400mK & Dhead < 600mK, start dilution module
    turn_on_still_heater = State(name = 'Turn on still heater', value = 15)
    #-----------------------------------------------------------------------------------------------------#
    # To speed up system warm up, turning on the heat switches to thermally connect the whole system & HPHs
    # Also turn off still heater? maybe keep it on to warm up still faster... It jsut wont cycle heilum bc not super fluid
    warm_up_system = State(name = 'Warming up system', value = 16, final = True)
    #-----------------------------------------------------------------------------------------------------#
    
   
    #-----------------------------------------------------------------------------------------------------#
    #                               Defining transitions between states
    #-----------------------------------------------------------------------------------------------------#
    # Normal DR cycle between states
    cycle = (
        wait_for_heat_switches_to_cool.to(turn_on_all_pumps, cond = 'wait_for_heat_switches_to_cool_cond')    | 
        turn_on_all_pumps.to(turn_on_4_heat_switches, cond = 'turn_on_all_pumps_cond')                        |
        turn_on_4_heat_switches.to(turn_on_3_heat_switches, cond = 'turn_on_4_heat_switches_cond')            |
        turn_on_3_heat_switches.to(start_cycle_A, cond = 'turn_on_3_heat_switches_cond')                      |

        start_cycle_A.to(get_pumps_warm_A, cond = 'start_cycle_A_cond')                                       |
        get_pumps_warm_A.to(get_pumps_hot_A, cond = 'get_pumps_warm_A_cond')                                  |
        get_pumps_hot_A.to(turn_on_heat_switch_4_A, cond ='get_pumps_hot_A_cond')                             |
        turn_on_heat_switch_4_A.to(turn_on_heat_switch_3_A, cond = 'turn_on_heat_switch_4_A_cond')            |
        turn_on_heat_switch_3_A.to(turn_on_still_heater, cond = 'dilution_cond')                              |
        turn_on_heat_switch_3_A.to(start_cycle_B, cond = 'turn_on_heat_switch_3_A_cond')                      |
        turn_on_still_heater.to(start_cycle_B, cond = 'still_heater_cond_B')                                  |

        start_cycle_B.to(get_pumps_warm_B, cond = 'start_cycle_B_cond')                                       |
        get_pumps_warm_B.to(get_pumps_hot_B, cond = 'get_pumps_warm_B_cond')                                  |
        get_pumps_hot_B.to(turn_on_heat_switch_4_B, cond = 'get_pumps_hot_B_cond')                            |
        turn_on_heat_switch_4_B.to(turn_on_heat_switch_3_B, cond = 'turn_on_heat_switch_4_B_cond')            |
        turn_on_heat_switch_3_B.to(turn_on_still_heater, cond = 'dilution_cond')                              |
        turn_on_heat_switch_3_B.to(start_cycle_A, cond = 'turn_on_heat_switch_3_B_cond')                      |           
        turn_on_still_heater.to(start_cycle_A, cond = 'still_heater_cond_A') |

        start_cycle_A.to(warm_up_system, cond = 'warm_up_system_cond')
        )

    #-----------------------------------------------------------------------------------------------------#
    #        Condition Functions - These determine if the state can transition to the next State          #
    #-----------------------------------------------------------------------------------------------------#
    
    def wait_for_heat_switches_to_cool_cond(self):
        # Goals
        # Check if Switches are below 7K and 4K is below 5Kelvin
        return self.check_temperature_criteria(self.current_state.id) 
    def turn_on_all_pumps_cond(self):
        # Goals
        # Check if all heads are below 6K
        return self.check_temperature_criteria(self.current_state.id) 
    
    def turn_on_4_heat_switches_cond(self):
        # Goals
        # Check if both 3Heads are below 1.55K
        return self.check_temperature_criteria(self.current_state.id) 
    
    def turn_on_3_heat_switches_cond(self):
        # Goals 
        # Check if 3Head PID timers passed their max elapsed time
        return self.timer_check()

    def start_cycle_A_cond(self):
        # Goals
        # Check if 4 & 3 Switch A are below 12K to start A pumps
        return self.check_temperature_criteria(self.current_state.id) 
        
    def get_pumps_warm_A_cond(self):
        # Goals
        # Checks timers for warming 4PUMPA and 3PUMPA            
        return self.timer_check()
    
    def get_pumps_hot_A_cond(self):
        # Goals
        # Checks 4PUMPA timer to see if it has passed
        return self.timer_check()
    
    def turn_on_heat_switch_4_A_cond(self):
        # Goals
        # Check if 4HEADA < 1K and 3HEADA < 6K or if 3UPMPA elapsed time has passed
        if self.check_temperature_criteria(self.current_state.id):
            return True
        elif self.timer_check():
            return True
        else:
            return False
        
    def turn_on_heat_switch_3_A_cond(self):
        # Goals
        # Check if 3SWITCHA & 4SWITCHA timers have passed
        return self.timer_check()
    
    def start_cycle_B_cond(self):
        # Goals
        # Check if 4 & 3 Switch B are below 12K to start B pumps
        return self.check_temperature_criteria(self.current_state.id)

    def get_pumps_warm_B_cond(self):
        # Goals
        # Checks timers for warming 4PUMPB and 3PUMPB
        return self.timer_check()

    def get_pumps_hot_B_cond(self):
        # Goals
        # Checks 4PUMPB timer to see if it has passed
        return self.timer_check()

    def turn_on_heat_switch_4_B_cond(self):
        # Goals
        # Check if 4HEADB < 1K and 3HEADB < 6K or if 3UPMPB elapsed time has passed
        if self.check_temperature_criteria(self.current_state.id):
            return True
        elif self.timer_check():
            return True
        else:
            return False
    def turn_on_heat_switch_3_B_cond(self):
        # Goals
        # Check if 3SWITCHB & 4SWITCHB timers have passed
        return self.timer_check()
    
    def dilution_cond(self):
        # Goals
        # Check if Still temperature < 400mK and Dhead is < 600mK. If self.still_activated == True, then return False and dont go back into the still_heater state
        if self.still_activated == True:
            return False
        return self.check_temperature_criteria(self.current_state.id)
    
    def still_heater_cond_A(self):
        # Goals
        # Checks self.current_cycle to direct still_heater state to either start_cycle_A or start_cycle_B
        if self.current_cycle == 'A':
            return False
        elif self.current_cycle == 'B':
            return True

    def still_heater_cond_B(self):
        # Goals
        # Checks self.current_cycle to direct still_heater state to either start_cycle_A or start_cycle_B
        if self.current_cycle == 'A':
            return True
        elif self.current_cycle == 'B':
            return False
    def warm_up_system_cond(self):
        return False
    #-----------------------------------------------------------------------------------------------------#
    #        On-entry functions - These are the functions/tasks that are executed on entry into a state.  #
    #-----------------------------------------------------------------------------------------------------#  
    #   
    # Initial on_enter functions
    def on_enter_wait_for_heat_switches_to_cool(self):
        print('on_enter: Waiting for all heat switches < 12K & 4K stage < 5K', f'{self.log_time()}')
        self.run_per_state()
    def on_enter_turn_on_all_pumps(self):
        self.run_per_state()
        print('on_enter: Turning on all Pumps', f'{self.log_time()}')
    def on_enter_turn_on_4_heat_switches(self):
        self.run_per_state()
        print('on_enter: Turning on 4HeatSwitches A & B and Turning on 4PUMP A & B', f'{self.log_time()}')
    def on_enter_turn_on_3_heat_switches(self):
        self.run_per_state()
        print('on_enter: Turning on 3HeatSwitches A & B and Turning on 3PUMP A & B', f'{self.log_time()}')

    # Cycle A on_enter functions
    def on_enter_start_cycle_A(self):
        self.run_per_state()
        print('on_enter: Starting cycle He3 & He4 A cycle. Waiting for 4 & 3 SWITCH A < 12K', f'{self.log_time()}')
        self.current_cycle = 'A'
    def on_enter_get_pumps_warm_A(self):
        self.run_per_state()
        print('on_enter: Warming 4PUMPA & 3PUMPA to TOML set value', f'{self.log_time()}')
    def on_enter_get_pumps_hot_A(self):
        self.run_per_state()
        print('on_enter: Heating 4PUMPA & 3PUMPA to 45K', f'{self.log_time()}')
    def on_enter_turn_on_heat_switch_4_A(self):
        self.run_per_state()
        print('on_enter: Turning off 4PUMPA & turning on 4SWITCHA', f'{self.log_time()}\n')
        print('Waiting for 4HEADA < 1K and 3HEADA < 6K or if 3PUMPA elapsed time has passed', f'{self.log_time()}')
    def on_enter_turn_on_heat_switch_3_A(self):
        self.run_per_state()
        print('on_enter: Turning off 3PUMPA & turning on 3SWITCHA', f'{self.log_time()}\n')
        print('Checking if Still < 400mK and Dhead < 600mK', f'{self.log_time()}\n')
        print('or ...Waiting on 3SWITCHA elapse timers', f'{self.log_time()}')

    # Cycle B on_enter functions
    def on_enter_start_cycle_B(self):
        self.run_per_state()
        print('on_enter: Starting cycle He3 & He4 B cycle. Waiting for 4 & 3 SWITCH B < 12K ', f'{self.log_time()}')
        self.current_cycle = 'B'
    def on_enter_get_pumps_warm_B(self):
        self.run_per_state()
        print('on_enter: Warming 4PUMPB & 3PUMPB to TOML set value', f'{self.log_time()}')
    def on_enter_get_pumps_hot_B(self):
        self.run_per_state()
        print('on_enter: Heating 4PUMPB & 3PUMPB to 45K', f'{self.log_time()}')
    def on_enter_turn_on_heat_switch_4_B(self):
        self.run_per_state()
        print('on_enter: Turning off 4PUMPB & turning on 4SWITCHB', f'{self.log_time()}\n')
        print('Waiting for 4HEADB < 1K and 3HEADB < 6K or if 3PUMPB elapsed time has passed', f'{self.log_time()}')
    def on_enter_turn_on_heat_switch_3_B(self):
        self.run_per_state()
        print('on_enter: Turning off 3PUMPB & turning on 3SWITCHB', f'{self.log_time()}\n')
        print('Checking if Still < 400mK and Dhead < 600mK', f'{self.log_time()}\n')
        print('or... Waiting on 3SWITCHB elapse timers', f'{self.log_time()}')

    # Dilution on_enter function
    def on_enter_dilution(self):
        self.run_per_state()
        self.still_activated = True
        print('on_enter: Starting Dilution cycle. Turning on Still heater', f'{self.log_time()}')

    
    
      #-----------------------------------------------------------------------------------------------------#
    #                                   functions used for grabbing data                                  #
    #-----------------------------------------------------------------------------------------------------#

    # This grabs values of nested sensor_keys from a specific dictionary key
    def get_values_from_keys(self,key, sensor_keys, data_json):
        results = {}
        for sensor in sensor_keys:
            if sensor in data_json[f'{key}']:
                results[sensor] = data_json[f'{key}'][sensor]
        return results
    
    # This function will call the query metric server and return only the temperature of the sensors
    def get_temperature(self, sensors):
        data = self.monitor_client.get_metrics()
        return self.get_values_from_keys(sensor_keys = sensors, data_json = data, key = 'temperatures')
    
    # This function will call the query metric server and return only the heater values
    def get_heater_value(self, heater):
        data = self.monitor_client.get_metrics()
        return self.get_values_from_keys(sensor_keys= heater, data_json = data, key = 'heaters')

    #-----------------------------------------------------------------------------------------------------#
    #                      Action functions used to communicate with HAL.                                 #
    #-----------------------------------------------------------------------------------------------------#
    
    def set_heater_value(self, heater, value):
        response = self.hal_client.set_heater_value(f'{heater}', f'{value}')
        return response

    #-----------------------------------------------------------------------------------------------------#
    # Functions for starting and stopping the state machine
    #------------------------------------------------------------------------------------------------------#
    
    def non_async__turn_on_state_machine(self):
        # Starts state machine cycle
        self.cycle_running = True
        self.non_async_run_cycle()
    def non_async__turn_off_state_machine(self):    
        # Stops state machine cycle
        self.cycle_running = False
    def non_async_run_cycle(self):
            while self.cycle_running:
                # Update the temperatures inside the state machine
                self.update_fridge_state()
                # Run the state machine to check if we should move to the next state or not
                # If we move to a new state, update the list of currently-in-use PIDs
                self.cycle()
                # Scan through the list of currently-in-use PIDs and update their setpoints
                self.update_pid_values()
                time.sleep(self.state_machine_cycle_time)
    #------------------------------------------------------------------------------------------------------#
    #                                       Misc functions for the state machine
    #------------------------------------------------------------------------------------------------------#
    def update_fridge_state(self):
        fridge_data = self.monitor_client.get_metrics()
        self.fridge_state = fridge_data
        self.dt = self.fridge_timer()[1]
        self.fridge_timer.reset()

    def update_pid_values(self):
    # Update the setpoints of all PIDs in the list of currently-in-use PIDs
        #print('Updating PID values and sending to HAL')
        for pid_key in self.currently_in_use_pids:
            # Get current temperature of key
            current_value = self.fridge_state['temperatures'][pid_key]
            # Define state machine setpoint of key Ex 4PUMP @ 45K
            setpoint = self.currently_in_use_pids[pid_key]
            # Update PID setpoint
            self.pid[pid_key].setpoint = setpoint
            # Get new value from PID
            new_value = self.pid[pid_key](input_ = current_value, dt = self.dt)
            # Send new value to HAL
            #print('Current Temp =',current_value ,'Setting', pid_key, 'to', new_value, 'with dt =', self.dt)
            self.set_heater_value(heater = pid_key, value = new_value)

    # Checks temperature of fridge and returns True if all temperatures are below the threshold
    def check_temperature_criteria(self, state, excluded_variable = None):
        threshold_temps = self.sm_settings[f'{state}']['temperature_thresholds']
        current_temps = self.fridge_state['temperatures']
        # Filters current_temps to only include keys that are in threshold_temps
        filtered_current_data = {key: current_temps[key] for key in threshold_temps if key in current_temps}
        # Exclude a specific variable from the check
        if all(filtered_current_data[thermometer] < threshold_temps[thermometer] 
               for thermometer in threshold_temps if thermometer != excluded_variable):
            return True
        else:
            return False
    def check_temperature_criteria_sensor(self, sensor):
        # Get the temperature of the sensor
        current_temp = self.fridge_state['temperatures'][sensor]
        # Get the threshold for the sensor
        state = self.current_state.id
        threshold_temp = self.sm_settings[f'{state}']['temperature_thresholds'][sensor]
        # Check if the current temperature is below the threshold
        if current_temp < threshold_temp:
            return True
        else:
            return False
    # This will find the heaters and switches that contain 'pid_set_value' in their subkey. 
    # It is only called on entry to make self.currently_in_use_pids
    def find_sub_subkey(self, state, sub_subkey):
        """
        Find all subkeys and their values where the given sub-subkey exists in a nested dictionary.

        Parameters:
        - data (dict): The main dictionary with nested structures.
        - sub_subkey (str): The sub-subkey to search for.

        Returns:
        - dict: A dictionary of subkeys and their values where the sub-subkey exists.
        """
        data = self.sm_settings[f'{state}']
        result = {}

        for key, subkeys in data.items():
            # Ensure subkeys is a dictionary
            if isinstance(subkeys, dict):
                for subkey, subvalue in subkeys.items():
                    # Check if the subvalue is a dictionary and contains the sub-subkey
                    if isinstance(subvalue, dict) and sub_subkey in subvalue:
                        result[subkey] = subvalue[sub_subkey]

        return result
    #Makes timer objects
    def make_timers(self, dict_of_timers):
        for key, value in dict_of_timers.items():
            self.time_elapse_dict[key] = ElapsedTime(value)
            
    #Function for cycling through timers and checking if true
    def timer_check(self):
        for key in self.time_elapse_dict:
            #Check if time passed is true
            #EDIT ME I remove [0]
            if self.time_elapse_dict[key]()[0]:
                return True
        return False

    # Function used for keeping track of time in the state
    # EDIT ME
    def check_just_entered(self):
        if self.just_entered:
            self.just_entered = False
            self.enterStateTime = time.time()
            #Console / logging stuff
            return True
        else:
            return False

    def turn_off_not_in_use_pids(self):
        for key in self.pid:
            # Skip STILLHEATER if still_activated is True
            if key == "STILLHEATER" and self.still_activated:
                continue
            if key not in self.currently_in_use_pids.keys():
                print('Turning off', key)
                self.set_heater_value(key, 0)

    def run_per_state(self):
        # Resets time_elapse_dict
        self.time_elapse_dict = {}
        # Save STILLHEATER PID if it exists
        saved_pid_key = self.currently_in_use_pids.get('STILLHEATER')
        # Clear all PIDs
        self.currently_in_use_pids = {}
        # Restore STILLHEATER PID if it was saved
        if saved_pid_key is not None:
            self.currently_in_use_pids['STILLHEATER'] = saved_pid_key
        # Get new PIDs from state machine settings
        new_pids = self.find_sub_subkey(self.current_state.id, sub_subkey = 'pid_set_value')
        # Update the dictionary without overwriting STILLHEATER
        for key, value in new_pids.items():
            if key != 'STILLHEATER' or 'STILLHEATER' not in self.currently_in_use_pids:
                self.currently_in_use_pids[key] = value
        # Sets up elapse time objects
        self.dict_to_turn_into_timers = self.find_sub_subkey(self.current_state.id, sub_subkey = 'elapse_time_max')
        self.make_timers(self.dict_to_turn_into_timers)
        self.turn_off_not_in_use_pids()
        # EDIT ME 
        self.check_just_entered()

    def log_time(self):
        date = time.localtime()
        log_time = f'{date.tm_year}-{date.tm_mon}-{date.tm_mday} {date.tm_hour}:{date.tm_min}:{date.tm_sec}'
        return log_time

class ElapsedTime:
    def __init__(self, threshold_seconds=None):
        """
        Initialize the ElapsedTime object.
    

        Parameters:
        - threshold_seconds (float, optional): The time threshold in seconds.
          If None, threshold checks will not be performed.
        """
        self.threshold_seconds = threshold_seconds
        self.start_time = time.time()

    def reset(self):
        """
        Reset the timer.
        """
        self.start_time = time.time()

    def __call__(self):
        """
        Check if the elapsed time exceeds the optional threshold and return elapsed time.

        Returns:
        - (bool or None, float): A tuple containing:
            - bool: True if elapsed time exceeds the threshold, False otherwise. 
                    If no threshold is provided, this is None.
            - float: Elapsed time in seconds since the timer was started.
        """
        elapsed_time = time.time() - self.start_time

        # Check if threshold_seconds is provided
        if self.threshold_seconds is None:
            return None, elapsed_time
        
        return elapsed_time >= self.threshold_seconds, elapsed_time

    def time_remaining(self):
        """
        Get the remaining time until the threshold is reached.

        Returns:
        - float or None: Time remaining in seconds. 
                         If the threshold is exceeded, returns 0.
                         If no threshold is set, returns None.
        """
        if self.threshold_seconds is None:
            return None

        elapsed = time.time() - self.start_time
        return max(0, self.threshold_seconds - elapsed)



import json
import threading 
import time
import datetime
from http.server import ThreadingHTTPServer, BaseHTTPRequestHandler
import fridgeos.zmqhelper as zmqhelper
from fridgeos import HALClient

class S(BaseHTTPRequestHandler):
    """ Taken from https://gist.github.com/nitaku/10d0662536f37a087e1b """
    def do_GET(self):    
        self.send_response(200)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        sec_last_update = round(time.time() - self.server.json_dict['metadata']['last_update_time'],3)
        self.server.json_dict['metadata']['seconds_since_last_update'] = sec_last_update
        self.wfile.write(json.dumps(self.server.json_dict).encode(encoding='utf_8'))

class SimpleJSONhttpserver:
    def __init__(self, ip_address="localhost", port=8000):
        self.json_dict = {}
        self.httpd = ThreadingHTTPServer((ip_address, port), S)
        self.httpd.json_dict = self.json_dict
        self.thread = threading.Thread(target=self.httpd.serve_forever, args=())
        self.thread.start()

class MetricsServer:
    def __init__(self, cryostat_name, ip_address="localhost", port=8000):
        self.server = SimpleJSONhttpserver(ip_address=ip_address, port=port)
        self.server.json_dict['metadata'] = {'cryostat_name' : cryostat_name,
                                             'seconds_since_last_update' : -1,
                                             'last_update_time' : time.time(),
                                             'last_update_datetime' : str(datetime.datetime.now())}

    def update_time(self):
        self.server.json_dict['metadata']['last_update_time'] = time.time()
        self.server.json_dict['metadata']['last_update_datetime'] = str(datetime.datetime.now())
    
    def update_metric_values(self, metric_name, new_values_dict):
        self.server.json_dict[metric_name] = new_values_dict
        self.update_time()


class MonitorServer:
    def __init__(self,
                 cryostat_name,
                 http_port,
                 hal_ip,
                 hal_port,
                 min_update_period = 1,
        ):
        """ Create an HTTP server on port http_port that displays the metrics of
        the cryostat as a simple JSON. The monitor server will query the HAL
        server every min_update_period seconds for the temperatures/heater
        values/state
        """
        self.metrics_server = MetricsServer(cryostat_name = cryostat_name, 
                                            ip_address="0.0.0.0",
                                            port=http_port)
        self.hal_client = HALClient(ip = hal_ip, port = hal_port)
        self.min_update_period = min_update_period # seconds
        self.run()

    def update(self):
        # Get temperatures from HAL
        temperatures = self.hal_client.get_temperatures()
        self.metrics_server.update_metric_values(metric_name = 'temperatures',
                                                 new_values_dict = temperatures)

        # Get heater values from HAL
        heater_values = self.hal_client.get_heater_values()
        self.metrics_server.update_metric_values(metric_name = 'heaters',
                                                 new_values_dict = heater_values)

        # Get heater max values from HAL
        heater_max_values = self.hal_client.get_heater_max_values()
        self.metrics_server.update_metric_values(metric_name = 'heater_max_values',
                                                 new_values_dict = heater_max_values)

        # FIXME implement state update

    def run(self):
        print('Starting monitor server')
        while True:
            try:
                time_start = time.time()
                self.update()
                while time.time() - time_start < self.min_update_period:
                    time.sleep(0.01)
            except KeyboardInterrupt:
                print('Stopping monitor server')
                break
            except Exception as e:
                print(e)
                time.sleep(1)



from fridgeos.hal.server import HALServer
from fridgeos.hal.client import HALClient
from fridgeos.monitor.server import MonitorServer
from fridgeos.monitor.client import MonitorClient
from fridgeos.scraper.scraper import Scraper, PostgresUploader
# from fridgeos.statemachine import single_shot_1k
# from fridgeos.statemachine import continous_dr



import serial
import numpy as np

class SIM922:
    """Python class for SRS SIM922 quad thermometer inside a SIM900
    mainframe, written by Adam McCaughan"""
    def __init__(self, serial_address, sim900port, channel):
        self.serial = serial.Serial(serial_address, timeout = 1, baudrate = 115200)
        self.sim900port = sim900port
        self.channel = channel
        self.serial.send_break()

    def read(self):
        return self.serial.readline().strip().decode('utf-8')

    def write(self, string):
        write_string = string + '\r\n'
        self.serial.write(write_string.encode('utf-8'))

    def query(self, string):
        self.write(string)
        return self.read()

    def reset(self):
        self.write_simport('*RST')

    def identify(self):
        return self.query_simport('*IDN?')

    def close(self):
        return self.serial.close()

    def write_simport(self, message):
        self.write('CONN %d,"xyz"' % self.sim900port)
        self.write(message)
        self.write('xyz')

    def read_simport(self):
        self.write('CONN %d,"xyz"' % self.sim900port)
        data = self.read()
        self.write('xyz')
        return data

    def query_simport(self, message):
        self.write('CONN %d,"xyz"' % self.sim900port)
        reply = self.query(message)
        self.write('xyz')
        return reply

    def read_temperature(self):
        # In a string, %0.4e converts a number to scientific notation
        self.write('CONN %d,"xyz"' % self.sim900port)
        v = float(self.query('TVAL? %s' % self.channel))
        self.write('xyz')
        return v

# s = SIM922('/dev/ttyUSB0', 1)
# print(s.identify())